"""
Generalization Building Model
=============================

This method is to create random simlpe buildings to train a policy that generalizate
the control of different devices in a building.
"""
import os
import json
import numpy as np
from typing import Dict, Any
from eprllib.EpisodeFunctions.EpisodeFunctions import EpisodeFunction
from eprllib.Tools.Utils import building_dimension, inertial_mass, u_factor, random_weather_config

class GeneralBuilding(EpisodeFunction):
    def __init__(
        self, episode_fn_config:Dict[str,Any]
    ):
        super().__init__(episode_fn_config)
        self.epjson_files_folder_path = episode_fn_config['epjson_files_folder_path']
    
    def get_episode_config(self, env_config:Dict[str,Any]) -> Dict[str,Any]:
        """
        This method define the properties of the episode. Changing some properties as weather or 
        Run Time Period, and defining others fix properties as volumen or window area relation.
        
        Return:
            dict: The method returns the env_config with modifications.
        """
        epjson_path: str = env_config['epjson_path']
        # Establish the epJSON Object, it will be manipulated to modify the building model.
        with open(epjson_path) as file:
            epJSON_object: dict = json.load(file)
        
        # == BUILDING ==
        # The building volume is V=h(high)*w(weiht)*l(large) m3
        h = 3
        w = (10-3) * np.random.random_sample() + 3
        l = (10-3) * np.random.random_sample() + 3
        
        # Calculate the aspect relation
        env_config['building_properties']['building_area'] = w*l
        env_config['building_properties']['aspect_ratio'] = w/l
        
        # Change the dimension of the building and windows
        window_area_relation = []
        env_config['building_properties']['window_area_relation_north'] = (0.9-0.05) * np.random.random_sample() + 0.05
        env_config['building_properties']['window_area_relation_east'] = (0.9-0.05) * np.random.random_sample() + 0.05
        env_config['building_properties']['window_area_relation_south'] = (0.9-0.05) * np.random.random_sample() + 0.05
        env_config['building_properties']['window_area_relation_west'] = (0.9-0.05) * np.random.random_sample() + 0.05
        window_area_relation.append(env_config['building_properties']['window_area_relation_north'])
        window_area_relation.append(env_config['building_properties']['window_area_relation_east'])
        window_area_relation.append(env_config['building_properties']['window_area_relation_south'])
        window_area_relation.append(env_config['building_properties']['window_area_relation_west'])
        window_area_relation = np.array(window_area_relation)
        building_dimension(epJSON_object, h, w, l,window_area_relation)
        
        # Define the type of construction (construction properties for each three layers)
        # Walls
        # Exterior Finish
        # Se toman las propiedades en el rango de valores utilizados en BOpt
        epJSON_object["Material"]["wall_exterior"]["thickness"] = (0.1017-0.0095) * np.random.random_sample() + 0.0095
        epJSON_object["Material"]["wall_exterior"]["conductivity"] = (0.7934-0.0894) * np.random.random_sample() + 0.0894
        epJSON_object["Material"]["wall_exterior"]["density"] = (1762.3-177.822) * np.random.random_sample() + 177.822
        epJSON_object["Material"]["wall_exterior"]["specific_heat"] = (1172.37-795.53) * np.random.random_sample() + 795.53
        epJSON_object["Material"]["wall_exterior"]["thermal_absorptance"] = (0.97-0.82) * np.random.random_sample() + 0.82
        epJSON_object["Material"]["wall_exterior"]["solar_absorptance"] = epJSON_object["Material"]["wall_exterior"]["visible_absorptance"] = (0.89-0.3) * np.random.random_sample() + 0.3
        # Inter layers
        # Se realizó un equivalente de los diferentes materiales utilizados en BOpt para resumirlos en una sola capa equivalente.
        epJSON_object["Material"]["wall_inter"]["thickness"] = (0.29-0.1524) * np.random.random_sample() + 0.1524
        epJSON_object["Material"]["wall_inter"]["conductivity"] = (0.8656-0.0474) * np.random.random_sample() + 0.0474
        epJSON_object["Material"]["wall_inter"]["density"] = (1822.35-84.57) * np.random.random_sample() + 84.57
        epJSON_object["Material"]["wall_inter"]["specific_heat"] = (1214.24-852.57) * np.random.random_sample() + 852.57
        epJSON_object["Material"]["wall_inter"]["thermal_absorptance"] = 0.9
        epJSON_object["Material"]["wall_inter"]["solar_absorptance"] = epJSON_object["Material"]["wall_inter"]["visible_absorptance"] = (0.71-0.65) * np.random.random_sample() + 0.65
        
        # Ceiling/Roof (exterior, inter layers)
        # Exterior Finish
        epJSON_object["Material"]["roof_exterior"]["thickness"] = (0.01906-0.0005) * np.random.random_sample() + 0.0005
        epJSON_object["Material"]["roof_exterior"]["conductivity"] = (50.04-0.1627) * np.random.random_sample() + 0.1627
        epJSON_object["Material"]["roof_exterior"]["density"] = (7801.75-1121.4) * np.random.random_sample() + 1121.4
        epJSON_object["Material"]["roof_exterior"]["specific_heat"] = (1465.46-460.57) * np.random.random_sample() + 460.57
        epJSON_object["Material"]["roof_exterior"]["thermal_absorptance"] = (0.95-0.88) * np.random.random_sample() + 0.88
        epJSON_object["Material"]["roof_exterior"]["solar_absorptance"] = epJSON_object["Material"]["roof_exterior"]["visible_absorptance"] = (0.93-0.6) * np.random.random_sample() + 0.6
        # Inter layers
        epJSON_object["Material"]["roof_inner"]["thickness"] = (0.1666-0.0936) * np.random.random_sample() + 0.0936
        epJSON_object["Material"]["roof_inner"]["conductivity"] = 0.029427
        epJSON_object["Material"]["roof_inner"]["density"] = 32.04
        epJSON_object["Material"]["roof_inner"]["specific_heat"] = 1214.23
        epJSON_object["Material"]["roof_inner"]["thermal_absorptance"] = 0.9
        epJSON_object["Material"]["roof_inner"]["solar_absorptance"] = epJSON_object["Material"]["roof_inner"]["visible_absorptance"] = 0.7

        # Internal Mass layer (interior layer)
        epJSON_object["Material"]["wall_inner"]["thickness"] = epJSON_object["Material"]["roof_inner"]["thickness"] = (0.0318-0.0127) * np.random.random_sample() + 0.0127
        epJSON_object["Material"]["wall_inner"]["conductivity"] = epJSON_object["Material"]["roof_inner"]["conductivity"] = 0.1602906
        epJSON_object["Material"]["wall_inner"]["density"] = epJSON_object["Material"]["roof_inner"]["density"] = 801
        epJSON_object["Material"]["wall_inner"]["specific_heat"] = epJSON_object["Material"]["roof_inner"]["specific_heat"] = 837.4
        epJSON_object["Material"]["wall_inner"]["thermal_absorptance"] = epJSON_object["Material"]["roof_inner"]["thermal_absorptance"] = 0.9
        epJSON_object["Material"]["wall_inner"]["solar_absorptance"] = epJSON_object["Material"]["wall_inner"]["visible_absorptance"] = epJSON_object["Material"]["roof_inner"]["solar_absorptance"] = epJSON_object["Material"]["roof_inner"]["visible_absorptance"] = 0.6

        # Windows
        # Change the window thermal properties
        epJSON_object['WindowMaterial:SimpleGlazingSystem']['WindowMaterial']['u_factor'] = (4.32-0.9652) * np.random.random_sample() + 0.9652
        epJSON_object['WindowMaterial:SimpleGlazingSystem']['WindowMaterial']['solar_heat_gain_coefficient'] = (0.67-0.26) * np.random.random_sample() + 0.26
        
        # The internal thermal mass is modified.
        for key in [key for key in epJSON_object["InternalMass"].keys()]:
            epJSON_object["InternalMass"][key]["surface_area"] = np.random.randint(10,40)
        
        # The total inertial thermal mass is calculated.
        # env_config['building_properties']['inercial_mass'] = inertial_mass(epJSON_object)
        
        # The global U factor is calculated.
        # env_config['building_properties']['construction_u_factor'] = u_factor(epJSON_object)
        
        # The limit capacity of bouth cooling and heating are changed.
        E_cool_ref = env_config['building_properties']['E_cool_ref'] = (3000 - 100)*np.random.random_sample() + 100
        E_heat_ref = env_config['building_properties']['E_heat_ref'] = (3000 - 100)*np.random.random_sample() + 100
        HVAC_names = [key for key in epJSON_object["ZoneHVAC:IdealLoadsAirSystem"].keys()]
        number_of_timesteps_per_hour = epJSON_object['Timestep']['Timestep 1']['number_of_timesteps_per_hour']
        for hvac in range(len(HVAC_names)):
            epJSON_object["ZoneHVAC:IdealLoadsAirSystem"][HVAC_names[hvac]]["maximum_sensible_heating_capacity"] = E_heat_ref
            epJSON_object["ZoneHVAC:IdealLoadsAirSystem"][HVAC_names[hvac]]["maximum_total_cooling_capacity"] = E_cool_ref
        if env_config.get('reward_fn_config', False):
            if env_config['reward_fn_config'].get('cooling_energy_ref', False):
                env_config['reward_fn_config']['cooling_energy_ref'] = E_cool_ref*number_of_timesteps_per_hour*3600
            if env_config['reward_fn_config'].get('heating_energy_ref', False):
                env_config['reward_fn_config']['heating_energy_ref'] = E_heat_ref*number_of_timesteps_per_hour*3600
        
                
        # Select the schedule file for loads
        env_config = random_weather_config(env_config)
        
        # The new modify epjson file is writed in the results folder created by RLlib
        # If don't exist, reate a folder call 'models' into env_config['episode_config']['epjson_files_folder_path']
        if not os.path.exists(f"{self.epjson_files_folder_path}/models"):
            os.makedirs(f"{self.epjson_files_folder_path}/models")
        
        env_config["epjson_path"] = f"{self.epjson_files_folder_path}/models/model-{env_config['episode']:08}-{os.getpid():05}.epJSON"
        with open(env_config["epjson_path"], 'w') as fp:
            json.dump(epJSON_object, fp, sort_keys=False, indent=4)
            # The new modify epjson file is writed.
        
        return env_config
    


"""
On-Off Air Conditioner Control Agent
====================================

This module contains the On-Off Air Conditioner Control agent that may 
involve in a dwelling.

The base class is :class:`~eprllib.Agents.ConventionalAgent.ConventionalAgent`.
"""

from typing import Dict, Any
from eprllib.Agents.ConventionalAgent import ConventionalAgent

class OnOffAirConditionerControl(ConventionalAgent):
    def __init__(
        self,
        config: Dict[str, Any],
    ):
        """
        Control of the Air Conditioner with an On-Off strategy.

        Args:
            config (Dict[str, Any]): This dictionary contains the configuration of the agent.
            It must contain the keys 'SP_temp', 'dT_up', 'dT_dn' and the keys that correspond to the
            temperature ('Ti') variable in the EnergyPlus model.
        """
        super().__init__(config)
        
    def compute_single_action(self, infos:Dict, prev_action) -> int:
        """
        This method allows an binary operation of an Air Conditioner througt fixed rule
        based control.

        Args:
            infos (Dict): Dictionary that contains the observation of the environment needed to
            implement the control policy. In this case, the dictionary must contain the keys and 
            values corresponding to the variables of zone temperature in the EnergyPlus model.
            prev_action (float): Previous action applied by the agent in the environment.

        Returns:
            int: Return the action to be applied in the EnergyPlus model environment. (0 if Off
            and 1 if On. If Any Error apears, return -1. This could be used as a flag.
        """
        
        # Se obtiene la configuración
        SP_temp = self.config['SP_temp']
        dT_up = self.config['dT_up']
        dT_dn = self.config['dT_dn']
        
        Ti = infos[self.config['Ti']]
        
        if Ti >= (SP_temp + dT_up):
            action_aa = 1

        elif Ti <= (SP_temp - dT_dn):
            action_aa = 0

        elif Ti < (SP_temp + dT_up) and Ti > (SP_temp - dT_dn):
            action_aa = prev_action

        else:
            print(f"On-Off Air Conditioner control fail. The policy apply was configured as:\n{self.config}\The infos dictionary was:\n{infos}\nand the previous action was:\n{prev_action}")
            action_aa = -1
        
        return action_aa


"""
Environment Configuration
=========================

This module contain the class and methods used to configure the environment.
"""

from typing import Optional, List, Dict, Tuple, Any
from eprllib.ActionFunctions.ActionFunctions import ActionFunction
from eprllib.RewardFunctions.RewardFunctions import RewardFunction
from eprllib.EpisodeFunctions.EpisodeFunctions import EpisodeFunction

def env_config_to_dict(EnvConfig) -> Dict:
    """
    Convert an EnvConfig object into a dict before to be used in the env_config parameter of RLlib environment config.
    """
    return vars(EnvConfig)

class EnvConfig:
    def __init__(self):
        """
        This is the main object that it is used to relate the EnergyPlus model and the RLlib policy training execution.
        """
        # generals
        self.epjson_path: str = ''
        self.epw_path: str = ''
        self.output_path: str = ''
        self.ep_terminal_output: bool = True
        self.timeout: float = 10.0

        # agents
        self.agents_config: Dict[str,Dict[str,Any]] = {}

        # observations
        self.ep_environment_variables: List|bool = False
        self.ep_thermal_zones_variables: List|bool = False
        self.ep_object_variables: Dict[str,Dict[str,Tuple[str,str]]]|bool = False
        self.ep_meters: List|bool = False
        self.time_variables: List|bool = False
        self.weather_variables: List|bool = False
        self.infos_variables: Dict[str,List]|bool = False
        self.no_observable_variables: Dict[str,List]|bool = False
        self.use_actuator_state: bool = False
        self.use_agent_indicator: bool = False
        self.use_agent_type: bool = False
        self.use_building_properties: bool = False
        self.buildig_properties: Dict[str,Dict[str,float]] = {}
        self.use_one_day_weather_prediction: bool = False

        # actions
        self.action_fn: ActionFunction = ActionFunction

        # rewards
        self.reward_fn: RewardFunction = RewardFunction

        # functionalities
        self.cut_episode_len: int = 1
        self.episode_fn: EpisodeFunction = EpisodeFunction
    
    def generals(
        self, 
        epjson_path:str,
        epw_path:str,
        output_path:str,
        ep_terminal_output:Optional[bool] = True,
        timeout:Optional[float] = 10.0
        ):
        """
        This method is used to modify the general configuration of the environment.

        Args:
            epjson_path (str): The path to the EnergyPlus model in the format of epJSON file.
            epw_path (str): The path to the EnergyPlus weather file in the format of epw file.
            output_path (str): The path to the output directory for the EnergyPlus simulation.
            ep_terminal_output (bool): For dubugging is better to print in the terminal the outputs 
            of the EnergyPlus simulation process.
            timeout (float): timeout define the time that the environment wait for an observation 
            and the time that the environment wait to apply an action in the EnergyPlus simulation. 
            After that time, the episode is finished. If your environment is time consuming, you 
            can increase this limit. By default the value is 10 seconds.
        """
        self.epjson_path = epjson_path
        self.epw_path = epw_path
        self.output_path = output_path
        self.ep_terminal_output = ep_terminal_output
        self.timeout = timeout
        
    def agents(
        self, 
        agents_config:Dict[str,Dict[str,Any]]
        ):
        """
        This method is used to modify the agents configuration of the environment.

        Args:
            agents_config (Dict[str,Dict[str,Any]]): This dictionary contain the names of the agents 
            involved in the environment. The mandatory components of the agent are: ep_actuator_configuration, 
            thermal_zone, actuator_type, agent_indicator.
        """
        self.agents_config = agents_config
    
    def observations(
        self,
        ep_environment_variables: List[str]|bool = False,
        ep_thermal_zones_variables: List[str]|bool = False,
        ep_object_variables: Dict[str,Dict[str,Tuple[str,str]]]|bool = False,
        ep_meters: List[str]|bool = False,
        time_variables: List[str]|bool = False,
        weather_variables: List[str]|bool = False,
        infos_variables: Dict[str,List[str]]|bool = False,
        no_observable_variables: Dict[str,List[str]]|bool = False,
        use_actuator_state: Optional[bool] = False,
        use_agent_indicator: Optional[bool] = False,
        use_agent_type: Optional[bool] = False,
        use_building_properties: Optional[bool] = False,
        buildig_properties: Optional[Dict[str,Dict[str,float]]] = {},
        use_one_day_weather_prediction: Optional[bool] = False,
        ):
        """
        This method is used to modify the observations configuration of the environment.

        Args:
            use_actuator_state (bool): define if the actuator state will be used as an observation for the agent.
            use_agent_indicator (bool): define if agent indicator will be used as an observation for the agent. 
            This is recommended True for muilti-agent usage and False for single agent case.
            use_agent_type (bool): define if the agent/actuator type will be used. This is recommended for different 
            types of agents actuating in the same environment.
            use_building_properties (bool): # define if the building properties will be used as an observation for 
            the agent. This is recommended if different buildings/thermal zones will be used with the same policy.
            buildig_properties (Dict[str,Dict[str,float]]): # The episode config define important aspects about the 
            building to be simulated in the episode.
            use_one_day_weather_prediction (bool): We use the internal variables of EnergyPlus to provide with a 
            prediction of the weathertime ahead. The variables to predict are:
            
            * Dry Bulb Temperature in °C with squer desviation of 2.05 °C, 
            * Relative Humidity in % with squer desviation of 20%, 
            * Wind Direction in degree with squer desviation of 40°, 
            * Wind Speed in m/s with squer desviation of 3.41 m/s, 
            * Barometric pressure in Pa with a standart deviation of 1000 Pa, 
            * Liquid Precipitation Depth in mm with desviation of 0.5 mm.
            
            This are predicted from the next hour into the 24 hours ahead defined.
            ep_environment_variables (List[str]):
            ep_thermal_zones_variables (List[str]): 
            ep_object_variables (Dict[str,Dict[str,Tuple[str,str]]]): 
            ep_meters (List[str]): names of meters from EnergyPlus to observe.
            time_variables (List[str]): The time variables to observe in the EnergyPlus simulation. The format is a 
            list of the names described in the EnergyPlus epJSON format documentation (https://energyplus.readthedocs.io/en/latest/schema.html) 
            related with temporal variables. All the options are listed bellow.
            weather_variables (List[str]): The weather variables are related with weather values in the present timestep 
            for the agent. The following list provide all the options avialable. To weather predictions see the 'weather_prob_days' 
            config that is follow in this file.
            infos_variables (Dict[str,List[str]]): The information variables are important to provide information for the 
            reward function. The observation is pass trough the agent as a NDArray but the info is a dictionary. In this 
            way, we can identify clearly the value of a variable with the key name. All the variables used in the reward 
            function must to be in the infos_variables list. The name of the variables must to corresponde with the names 
            defined in the earlier lists.
            no_observable_variables (Dict[str,List[str]]): There are occasions where some variables are consulted to use in 
            training but are not part of the observation space. For that variables, you can use the following  list. An strategy, 
            for example, to use the Fanger PPD value in the reward function but not in the observation space is to aggregate the 
            PPD into the 'infos_variables' and in the 'no_observable_variables' list.
        """
        # TODO: Al least one variable must to be defined.
        self.use_actuator_state = use_actuator_state
        self.use_agent_indicator = use_agent_indicator
        self.use_agent_type = use_agent_type
        self.use_building_properties = use_building_properties
        self.buildig_properties = buildig_properties
        self.use_one_day_weather_prediction = use_one_day_weather_prediction
        self.ep_environment_variables = ep_environment_variables
        self.ep_thermal_zones_variables = ep_thermal_zones_variables
        self.ep_object_variables = ep_object_variables
        self.ep_meters = ep_meters
        self.time_variables = time_variables
        self.weather_variables = weather_variables
        self.infos_variables = infos_variables
        self.no_observable_variables = no_observable_variables
    
    def actions(
        self,
        action_fn: ActionFunction = ActionFunction
        ):
        """
        This method is used to modify the actions configuration of the environment.
        
        Args:
            action_fn (ActionFunction): In the definition of the action space, usualy is use the discrete form of the 
            gym spaces. In general, we don't use actions from 0 to n directly in the EnergyPlus simulation. With the 
            objective to transform appropiately the discret action into a value action for EP we define the action_fn. 
            This function take the arguments agent_id and action. You can find examples in eprllib.ActionFunctions.
        """
        self.action_fn = action_fn

    def rewards(
        self,
        reward_fn: RewardFunction = RewardFunction,
        ):
        """
        This method is used to modify the rewards configuration of the environment.

        Args:
            reward_fn (RewardFunction): The reward funtion take the arguments EnvObject (the GymEnv class) and the infos 
            dictionary. As a return, gives a float number as reward. See eprllib.RewardFunctions for examples.
        """
        self.reward_fn = reward_fn

    def functionalities(
        self,
        episode_fn: EpisodeFunction = EpisodeFunction,
        cut_episode_len: int = 0,
        ):
        """
        This method configure special functions to improve the use of eprllib.

        Args:
            episode_fn (): This method define the properties of the episode, taking the env_config dict and returning it 
            with modifications.
            episode_config (Dict): NotDescribed
            cut_episode_len (int): Sometimes is useful to cut the simulation RunPeriod into diferent episodes. By default, 
            an episode is a entire RunPeriod EnergyPlus simulation. If you set the 'cut_episode_len' in 1 (day) you will 
            truncate the, for example, annual simulation into 365 episodes. If ypu set to 0, no cut will be apply.
        """
        self.episode_fn = episode_fn
        self.cut_episode_len = cut_episode_len



"""
Window Opening Control Agent
============================

This module contains the Window Opening Control agent that may 
involve in a dwelling.

The base class is :class:`~eprllib.Agents.ConventionalAgent.ConventionalAgent`.
"""

from typing import Dict, Any
from eprllib.Agents.ConventionalAgent import ConventionalAgent

class WindowOpeningControl(ConventionalAgent):
    def __init__(
        self,
        config: Dict[str, Any],
    ):
        """
        Control of the openings in windows.

        Args:
            config (Dict[str, Any]): This dictionary contains the configuration of the agent.
            It must contain the keys 'SP_temp', 'dT_up', 'dT_dn' and the keys that correspond to the
            indoor and outdoor temperatures ('Ti' and 'To' respectively) variables in the 
            EnergyPlus model.
        """
        super().__init__(config)
        
    def compute_single_action(self, infos:Dict, prev_action) -> int:
        """
        This method allows an binary operation of the opening of a window througt fixed rule
        based control.

        Args:
            infos (Dict): Dictionary that contains the observation of the environment needed to
            implement the control policy. In this case, the dictionary must contain the keys and 
            values corresponding to the variables of indoor and outdoor temperatures in the 
            EnergyPlus model.
            prev_action (float): Previous action applied by the agent in the environment.

        Returns:
            int: Return the action to be applied in the EnergyPlus model environment. (0 if the  
            window is close and 1 if the window is open. If Any Error apears, return -1. This 
            could be used as a flag.
        """
        # Se obtiene la configuración
        SP_temp = self.config['SP_temp']
        dT_up = self.config['dT_up']
        dT_dn = self.config['dT_dn']
        
        Ti = infos[self.config['Ti']]
        To = infos[self.config['To']]
        
        if Ti >= (SP_temp + dT_up):
            if Ti > To:
                action_v = 1
            else:
                action_v = 0

        elif Ti <= (SP_temp - dT_dn):
            if Ti > To:
                action_v = 0
            else:
                action_v = 1

        elif Ti < (SP_temp + dT_up) and Ti > (SP_temp - dT_dn):
            action_v = prev_action

        else:
            print(f"Window opening control fail. The policy apply was configured as:\n{self.config}\The infos dictionary was:\n{infos}\nand the previous action was:\n{prev_action}")
            action_v = -1
        
        return action_v


"""
Environments
============

The environments in RLlib can be implemented as a simple-agent, multi-agent or
external environment. In eprllib we use a multi-agent aproach, allowing to run 
multiples agents in the environment, but also a single agent if only one is specify.

The standard configuration use a policy with fully shared parameters, but in the next
versions we hope to add flexibility to the policy.

You can configure the environment with the :class:`~eprllib.Env.EnvConfig.EnvConfig` class .
"""



"""
Reward Functions
================

This module contain the reward functions used in the reinforcement learning and 
implemented in eprllib.

You can implement your own reward function by creating a new class that inherits from
:class:`~eprllib.RewardFunctions.RewardFunction` and overriding the `calculate_reward` method.
"""



"""
Multi-Agent Environment
=======================

The multi-agent aproach allow to run multiples agents in the environment, but also
a single agent.

The standard configuration use a policy with fully shared parameters.

You can configure the environment with the :class:`~eprllib.Env.EnvConfig.EnvConfig` class .
"""



"""
Window Shade Control Agent
==========================

This module contains the Window Shade Control agent that may 
involve in a dwelling.

The base class is :class:`~eprllib.Agents.ConventionalAgent.ConventionalAgent`.
"""

from typing import Dict, Any
from eprllib.Agents.ConventionalAgent import ConventionalAgent

class WindowShadeControl(ConventionalAgent):
    def __init__(
        self,
        config: Dict[str, Any],
    ):
        """
        Control of the shadows in windows.

        Args:
            config (Dict[str, Any]): This dictionary contains the configuration of the agent.
            It must contain the keys 'SP_temp', 'dT_up', 'dT_dn' and the keys that correspond to the
            temperature and solar radiation ('Ti' and 'Bw' respectively)variables in the 
            EnergyPlus model.
        """
        super().__init__(config)
        
    def compute_single_action(self, infos:Dict, prev_action:float) -> int:
        """
        This method allows an binary operation of a shadow (blind or shade) througt fixed rule
        based control.

        Args:
            infos (Dict): Dictionary that contains the observation of the environment needed to
            implement the control policy. In this case, the dictionary must contain the keys and 
            values corresponding to the variables of temperature and solar radiation in the EnergyPlus
            model.
            prev_action (float): Previous action applied by the agent in the environment.

        Returns:
            int: Return the action to be applied in the EnergyPlus model environment. (0 if must to not apply
            the shadow and 1 if the shadow must be apply. If Any Error apears, return -1. This could be used
            as a flag.
        """
        # Se obtiene la configuración
        SP_temp = self.config['SP_temp']
        dT_up = self.config['dT_up']
        dT_dn = self.config['dT_dn']
        
        Ti = infos[self.config['Ti']]
        Bw = infos[self.config['Bw']]
        
        #Control de la persiana
        if Ti >= (SP_temp + dT_up) and Bw == 0:
            action_p = 0 #Abrir la persiana
        elif Ti >= (SP_temp + dT_up) and Bw > 0:
            action_p = 1 #Cerrar la persiana
            
        elif Ti <= (SP_temp - dT_dn) and Bw == 0:
            action_p = 1
        elif Ti <= (SP_temp - dT_dn) and Bw > 0:
            action_p = 0
            
        elif Ti < (SP_temp + dT_up) and Ti > (SP_temp - dT_dn):
            action_p = prev_action

        else:
            print(f"Window shadow control fail. The policy apply was configured as:\n{self.config}\The infos dictionary was:\n{infos}\nand the previous action was:\n{prev_action}")
            action_p = -1
        
        return action_p


"""
Dual Set Point Thermostat
=========================

This module contains the implementation of the Dual Set Point Thermostat action function.
The action function transforms the action values from [0,1] to the appropriate range for the
actuator type (cooling, heating, or flow rate). The transformation is linear for cooling and
heating, and linear for flow rate.
"""

from typing import Dict, Any
from eprllib.ActionFunctions.ActionFunctions import ActionFunction

class DualSetPointThermostat(ActionFunction):
    def __init__(self, action_fn_config:Dict[str,Any]):
        """
        This class implements the Dual Set Point Thermostat action function.

        Args:
            action_fn_config (Dict[str,Any]): The configuration of the action function.
            It should contain the following keys: agents_type (Dict[str, int]): A dictionary 
            mapping agent names to their types (1 for cooling, 2 for heating, 3 for flow rate).
        """
        super().__init__(action_fn_config)
        self.agents_type: Dict[str, Any] = action_fn_config['agents_type']
    
    def transform_action(self, action:Dict[str,int]) -> Dict[str, float|int]:
        # create a dict to save the action per agent.
        action_transformed = {agent: 0. for agent in self.agents_type.keys()}
        for agent, type in self.agents_type.items():
            if type == 1: # Cooling
                # transfort value between [0,1] to [23,27]
                action_transformed[agent] = 23 + (action[agent]/10)*(27-23)
            elif type == 2: # Heating
                # transfort value between [0,1] to [18,22]
                action_transformed[agent] = 18 + (action[agent]/10)*(22-18)
            elif type == 3: # Flow Rate
                action_transformed[agent] = action[agent]/20
            else:
                raise ValueError('Actuator type not valid.')
        
        return action_transformed


"""
Action Functions
================

This module defines the functions that are used to perform actions on the
environment. The functions are used in the `EnvConfig` class to specify the
actions that can be performed on the environment.
"""



"""
Episode Functions Base Class
============================

This module contains the base class for the episode functions. This class is used to define the episode configuration
for the EnergyPlus environment.
"""
from typing import Dict, Any

class EpisodeFunction:
    """
    This class contains the methods to configure the episode in EnergyPlus with RLlib.
    """
    def __init__(
        self,
        episode_fn_config: Dict[str,Any] = {}
        ):
        self.episode_fn_config = episode_fn_config

    def get_episode_config(self, env_config: Dict[str,Any]) -> Dict[str,Any]:
        """
        This method returns the episode configuration for the EnergyPlus environment.

        Returns:
            Dict: The episode configuration.
        """
        return env_config


"""
Agents
======

This module contains classes for representing and manipulating agents in the
environment. The agents are responsible for taking actions in the environment
following a specified policy.
"""


"""# ENERGYPLUS RUNNER

This script contain the EnergyPlus Runner that execute EnergyPlus from its 
Python API in the version 23.2.0.
"""
import threading
import numpy as np
from queue import Queue
from time import sleep
from typing import Any, Dict, List, Optional, Set

from eprllib.ActionFunctions.ActionFunctions import ActionFunction
from eprllib.Env.MultiAgent.EnvUtils import (
    environment_variables, 
    thermal_zone_variables, 
    object_variables, 
    meters, 
    actuators, 
    EP_API_add_path,
)
# EnergyPlus Python API path adding
EP_API_add_path()
from pyenergyplus.api import EnergyPlusAPI
api = EnergyPlusAPI()

class EnergyPlusRunner:
    """
    This object have the particularity of `start` EnergyPlus, `_collect_obs` and `_send_actions` to
    send it trhougt queue to the EnergyPlus Environment thread.
    """
    def __init__(
        self,
        episode: int,
        env_config: Dict[str, Any],
        obs_queue: Queue,
        act_queue: Queue,
        infos_queue: Queue,
        _agent_ids: Set,
        _thermal_zone_ids: Set,
        ) -> None:
        """
        The object has an intensive interaction with EnergyPlus Environment script, exchange information
        between two threads. For a good coordination queue events are stablished and different canals of
        information are defined.

        Args:
            episode (int): Episode number.
            env_config (Dict[str, Any]): Environment configuration defined in the call to the EnergyPlus Environment.
            obs_queue (Queue): Queue object definition.
            act_queue (Queue): Queue object definition.
            infos_queue (Queue): Queue object definition.
        
        Return:
            None.
        """
        self.env_config = env_config
        
        # Asignation of variables.
        self.episode = episode
        self.obs_queue = obs_queue
        self.act_queue = act_queue
        self.infos_queue = infos_queue
        
        # The queue events are generated.
        self.obs_event = threading.Event()
        self.act_event = threading.Event()
        self.infos_event = threading.Event()
        self._agent_ids = _agent_ids
        self._thermal_zone_ids = _thermal_zone_ids
        # Variables to be used in this thread.
        self.energyplus_exec_thread: Optional[threading.Thread] = None
        self.energyplus_state: Any = None
        self.sim_results: int = 0
        self.initialized = False
        self.init_handles = False
        self.simulation_complete = False
        self.first_observation = True
        self.obs = {}
        self.infos = {}
        # create a variable to save the obs dict key to use in posprocess
        self.obs_keys = []
        self.infos_keys = []
        self.agent_actions = {}
        
        # Define the action transformer function
        self.action_fn: ActionFunction = self.env_config['action_fn']
        
        # Declaration of variables, meters and actuators to use in the simulation. Handles
        # are used in _init_handle method.
        self.variables, self.var_handles = environment_variables(self.env_config)
        self.thermal_zone_variables, self.thermal_zone_var_handles = thermal_zone_variables(self.env_config, self._thermal_zone_ids)
        self.object_variables, self.object_var_handles = object_variables(self.env_config, self._thermal_zone_ids)
        self.meters, self.meter_handles = meters(self.env_config)
        self.actuators, self.actuator_handles = actuators(self.env_config, self._agent_ids)
        
    def start(self) -> None:
        """
        This method inicialize EnergyPlus. First the episode is configurate, the calling functions
        established and the thread is generated here.
        """
        # Start a new EnergyPlus state (condition for execute EnergyPlus Python API).
        self.energyplus_state = api.state_manager.new_state()
        api.runtime.callback_begin_zone_timestep_after_init_heat_balance(self.energyplus_state, self._send_actions)
        api.runtime.callback_end_zone_timestep_after_zone_reporting(self.energyplus_state, self._collect_obs)
        api.runtime.set_console_output_status(self.energyplus_state, self.env_config['ep_terminal_output'])
        
        def _run_energyplus():
            """
            Run EnergyPlus in a non-blocking way with Threads.
            """
            cmd_args = self.make_eplus_args()
            print(f"running EnergyPlus with args: {cmd_args}")
            self.sim_results = api.runtime.run_energyplus(self.energyplus_state, cmd_args)
            self.simulation_complete = True
            
        self.energyplus_exec_thread = threading.Thread(
            target=_run_energyplus,
            args=()
        )
        # Here the thread is divide in two.
        self.energyplus_exec_thread.start()

    def _collect_obs(self, state_argument) -> None:
        """
        EnergyPlus callback that collects output variables, meters and actuator actions
        values and enqueue them to the EnergyPlus Environment thread.
        """
        # To not perform observations when the callbacks and the 
        # warming period are not complete.
        if not self._init_callback(state_argument) or self.simulation_complete:
            return
        
        hour = api.exchange.hour(state_argument)
        zone_time_step_number = api.exchange.zone_time_step_number(state_argument)
        
        # Register the state of the actuators
        self.agent_actions = {
            key: api.exchange.get_actuator_value(state_argument, handle)
            for key, handle
            in self.actuator_handles.items()
        }
        
        # thermal zone obs and infos dicts
        obs_tz = {thermal_zone: {} for thermal_zone in self._thermal_zone_ids}
        infos_tz = {thermal_zone: {} for thermal_zone in self._thermal_zone_ids}
        
        # Loop for each Thermal Zone conditioned.   
        for thermal_zone in self._thermal_zone_ids:
            
            if self.env_config['ep_environment_variables']:
                # Transform the list of names in Tuples
                obs_tz[thermal_zone].update(
                    {
                        key: api.exchange.get_variable_value(state_argument, handle)
                        for key, handle
                        in self.var_handles.items()
                    }
                )
            if self.env_config['ep_thermal_zones_variables']:
                # Transform the list of names in Tuples
                thermal_zone_thermal_zone_var_handles:Dict = self.thermal_zone_var_handles[thermal_zone]
                obs_tz[thermal_zone].update(
                    {
                        key: api.exchange.get_variable_value(state_argument, handle)
                        for key, handle
                        in thermal_zone_thermal_zone_var_handles.items()
                    }
                )
            if self.env_config['ep_object_variables']:
                # Transform the list of names in Tuples
                thermal_zone_object_var_handles:Dict = self.object_var_handles[thermal_zone]
                obs_tz[thermal_zone].update(
                    {
                        key: api.exchange.get_variable_value(state_argument, handle)
                        for key, handle
                        in thermal_zone_object_var_handles.items()
                    }
                )
            # add the meters if any.
            if self.env_config['ep_meters']:
                obs_tz[thermal_zone].update(
                    {
                        key: api.exchange.get_meter_value(state_argument, handle)
                        for key, handle
                        in self.meter_handles.items()
                    }
                )
            # Upgrade the observation with the building general properties
            if self.env_config['use_building_properties']:
                building_properties:Dict = self.env_config['building_properties'][thermal_zone]
                obs_tz[thermal_zone].update(building_properties)
        
            # Upgrade of the timestep observation with other variables.
            if self.env_config['time_variables']:
                time_variables_methods = {
                    'actual_date_time': api.exchange.actual_date_time(state_argument), # Gets a simple sum of the values of the date/time function. Could be used in random seeding.
                    'actual_time': api.exchange.actual_time(state_argument), # Gets a simple sum of the values of the time part of the date/time function. Could be used in random seeding.
                    'current_time': api.exchange.current_time(state_argument), # Get the current time of day in hours, where current time represents the end time of the current time step.
                    'day_of_month': api.exchange.day_of_month(state_argument), # Get the current day of month (1-31)
                    'day_of_week': api.exchange.day_of_week(state_argument), # Get the current day of the week (1-7)
                    'day_of_year': api.exchange.day_of_year(state_argument), # Get the current day of the year (1-366)
                    'holiday_index': api.exchange.holiday_index(state_argument), # Gets a flag for the current day holiday type: 0 is no holiday, 1 is holiday type #1, etc.
                    'hour': api.exchange.hour(state_argument), # Get the current hour of the simulation (0-23)
                    'minutes': api.exchange.minutes(state_argument), # Get the current minutes into the hour (1-60)
                    'month': api.exchange.month(state_argument), # Get the current month of the simulation (1-12)
                    'num_time_steps_in_hour': api.exchange.num_time_steps_in_hour(state_argument), # Returns the number of zone time steps in an hour, which is currently a constant value throughout a simulation.
                    'system_time_step': api.exchange.system_time_step(state_argument), # Gets the current system time step value in EnergyPlus. The system time step is variable and fluctuates during the simulation.
                    'year': api.exchange.year(state_argument), # Get the “current” year of the simulation, read from the EPW. All simulations operate at a real year, either user specified or automatically selected by EnergyPlus based on other data (start day of week + leap year option).
                    'zone_time_step': api.exchange.zone_time_step(state_argument), # Gets the current zone time step value in EnergyPlus. The zone time step is variable and fluctuates during the simulation.
                    'zone_time_step_number': api.exchange.zone_time_step_number(state_argument) # The current zone time step index, from 1 to the number of zone time steps per hour
                }
                time_variables_dict = {variable: time_variables_methods[variable] for variable in self.env_config['time_variables']}
                obs_tz[thermal_zone].update(time_variables_dict)
                
            if self.env_config['weather_variables']:
                weather_variables_methods = {
                    'is_raining': api.exchange.is_raining(state_argument), # Gets a flag for whether the it is currently raining. The C API returns an integer where 1 is yes and 0 is no, this simply wraps that with a bool conversion.
                    'sun_is_up': api.exchange.sun_is_up(state_argument), # Gets a flag for whether the sun is currently up. The C API returns an integer where 1 is yes and 0 is no, this simply wraps that with a bool conversion.
                    # Gets the specified weather data at the specified hour and time step index within that hour
                    'today_weather_albedo_at_time': api.exchange.today_weather_albedo_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_beam_solar_at_time': api.exchange.today_weather_beam_solar_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_diffuse_solar_at_time': api.exchange.today_weather_diffuse_solar_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_horizontal_ir_at_time': api.exchange.today_weather_horizontal_ir_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_is_raining_at_time': api.exchange.today_weather_is_raining_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_is_snowing_at_time': api.exchange.today_weather_is_snowing_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_liquid_precipitation_at_time': api.exchange.today_weather_liquid_precipitation_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_outdoor_barometric_pressure_at_time': api.exchange.today_weather_outdoor_barometric_pressure_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_outdoor_dew_point_at_time': api.exchange.today_weather_outdoor_dew_point_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_outdoor_dry_bulb_at_time': api.exchange.today_weather_outdoor_dry_bulb_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_outdoor_relative_humidity_at_time': api.exchange.today_weather_outdoor_relative_humidity_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_sky_temperature_at_time': api.exchange.today_weather_sky_temperature_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_wind_direction_at_time': api.exchange.today_weather_wind_direction_at_time(state_argument, hour, zone_time_step_number),
                    'today_weather_wind_speed_at_time': api.exchange.today_weather_wind_speed_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_albedo_at_time': api.exchange.tomorrow_weather_albedo_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_beam_solar_at_time': api.exchange.tomorrow_weather_beam_solar_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_diffuse_solar_at_time': api.exchange.tomorrow_weather_diffuse_solar_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_horizontal_ir_at_time': api.exchange.tomorrow_weather_horizontal_ir_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_is_raining_at_time': api.exchange.tomorrow_weather_is_raining_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_is_snowing_at_time': api.exchange.tomorrow_weather_is_snowing_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_liquid_precipitation_at_time': api.exchange.tomorrow_weather_liquid_precipitation_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_outdoor_barometric_pressure_at_time': api.exchange.tomorrow_weather_outdoor_barometric_pressure_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_outdoor_dew_point_at_time': api.exchange.tomorrow_weather_outdoor_dew_point_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_outdoor_dry_bulb_at_time': api.exchange.tomorrow_weather_outdoor_dry_bulb_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_outdoor_relative_humidity_at_time': api.exchange.tomorrow_weather_outdoor_relative_humidity_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_sky_temperature_at_time': api.exchange.tomorrow_weather_sky_temperature_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_wind_direction_at_time': api.exchange.tomorrow_weather_wind_direction_at_time(state_argument, hour, zone_time_step_number),
                    'tomorrow_weather_wind_speed_at_time': api.exchange.tomorrow_weather_wind_speed_at_time(state_argument, hour, zone_time_step_number)
                }
                weather_variables_dict = {variable: weather_variables_methods[variable] for variable in self.env_config['weather_variables']}
                obs_tz[thermal_zone].update(weather_variables_dict)
            
            # Weather prediction of 24 hours
            if self.env_config['use_one_day_weather_prediction']:
                weather_pred = {}
                # The list sigma_max contains the standard deviation of the predictions in the following order:
                #   - Dry Bulb Temperature in °C with squer desviation of 2.05 °C, 
                #   - Relative Humidity in % with squer desviation of 20%, 
                #   - Wind Direction in degree with squer desviation of 40°, 
                #   - Wind Speed in m/s with squer desviation of 3.41 m/s, 
                #   - Barometric pressure in Pa with a standart deviation of 1000 Pa, 
                #   - Liquid Precipitation Depth in mm with desviation of 0.5 mm.
                sigma_max = np.array([1.43178211, 4.47213595, 6.32455532, 1.84661853, 31.62, 0.707107])
                for h in range(24):
                    # For each hour, the sigma value goes from a minimum error of zero to the value listed in sigma_max following a linear function:
                    sigma = sigma_max * (h/23)
                    prediction_hour = hour+1 + h
                    if prediction_hour < 24:
                        weather_pred.update({
                            f'today_weather_liquid_precipitation_at_time_{prediction_hour}': max(0, np.random.normal(api.exchange.today_weather_liquid_precipitation_at_time(state_argument, prediction_hour, zone_time_step_number), sigma[5])),
                            f'today_weather_outdoor_barometric_pressure_at_time{prediction_hour}': max(0, np.random.normal(api.exchange.today_weather_outdoor_barometric_pressure_at_time(state_argument, prediction_hour, zone_time_step_number), sigma[4])),
                            f'today_weather_outdoor_dry_bulb_at_time{prediction_hour}': np.random.normal(api.exchange.today_weather_outdoor_dry_bulb_at_time(state_argument, prediction_hour, zone_time_step_number), sigma[0]),
                            f'today_weather_outdoor_relative_humidity_at_time{prediction_hour}': max(0, np.random.normal(api.exchange.today_weather_outdoor_relative_humidity_at_time(state_argument, prediction_hour, zone_time_step_number), sigma[1])),
                            f'today_weather_wind_direction_at_time{prediction_hour}': max(0, np.random.normal(api.exchange.today_weather_wind_direction_at_time(state_argument, prediction_hour, zone_time_step_number), sigma[2])),
                            f'today_weather_wind_speed_at_time{prediction_hour}': max(0, np.random.normal(api.exchange.today_weather_wind_speed_at_time(state_argument, prediction_hour, zone_time_step_number), sigma[3])),
                        })
                    else:
                        prediction_hour_t = prediction_hour - 24
                        weather_pred.update({
                            f'tomorrow_weather_liquid_precipitation_at_time{prediction_hour_t}': max(0, np.random.normal(api.exchange.tomorrow_weather_liquid_precipitation_at_time(state_argument, prediction_hour_t, zone_time_step_number), sigma[5])),
                            f'tomorrow_weather_outdoor_barometric_pressure_at_time{prediction_hour_t}': max(0, np.random.normal(api.exchange.tomorrow_weather_outdoor_barometric_pressure_at_time(state_argument, prediction_hour_t, zone_time_step_number), sigma[4])),
                            f'tomorrow_weather_outdoor_dry_bulb_at_time{prediction_hour_t}': np.random.normal(api.exchange.tomorrow_weather_outdoor_dry_bulb_at_time(state_argument, prediction_hour_t, zone_time_step_number), sigma[0]),
                            f'tomorrow_weather_outdoor_relative_humidity_at_time{prediction_hour_t}': max(0, np.random.normal(api.exchange.tomorrow_weather_outdoor_relative_humidity_at_time(state_argument, prediction_hour_t, zone_time_step_number), sigma[1])),
                            f'tomorrow_weather_wind_direction_at_time{prediction_hour_t}': max(0, np.random.normal(api.exchange.tomorrow_weather_wind_direction_at_time(state_argument, prediction_hour_t, zone_time_step_number), sigma[2])),
                            f'tomorrow_weather_wind_speed_at_time{prediction_hour_t}': max(0, np.random.normal(api.exchange.tomorrow_weather_wind_speed_at_time(state_argument, prediction_hour_t, zone_time_step_number), sigma[3])),
                        })
                obs_tz[thermal_zone].update(weather_pred)
                
                # Set the variables in the infos dict before to delete from the obs dict.
                if self.env_config['infos_variables']:
                    infos_tz[thermal_zone] = {variable: obs_tz[thermal_zone][variable] for variable in self.env_config['infos_variables'][thermal_zone]}
                # after infos assignation, delete not observable variables
                if self.env_config['no_observable_variables']:
                    for variable in self.env_config['no_observable_variables'][thermal_zone]:
                        del obs_tz[thermal_zone][variable]
                
                if self.first_observation:
                    self.obs_keys = [key for key in obs_tz[thermal_zone].keys()]
                    self.infos_keys = [key for key in infos_tz[thermal_zone].keys()]
                    if self.env_config['use_actuator_state']:
                        self.obs_keys.append('actuator_state')
                    if self.env_config['use_agent_indicator']:
                        self.obs_keys.append('agent_indicator')
                    if self.env_config['use_agent_type']:
                        self.obs_keys.append('agent_type')
                # ==FIN DEL LOOP DE OBSERVACIONES==
                
        # Ahora se tienen todas las observaciones e infos, una por cada zona térmica.
        
        # Se asignan observaciones y infos a cada agente.
        agents_obs = {agent: [] for agent in self._agent_ids}
        agents_infos = {agent: {} for agent in self._agent_ids}
        
        for agent in self._agent_ids:
            # Agent properties
            agent_thermal_zone = self.env_config['agents_config'][agent]['thermal_zone']
            agent_type = self.env_config['agents_config'][agent]['actuator_type']
            agent_indicator = self.env_config['agents_config'][agent]['agent_indicator']
            
            # Agent infos asignation
            agents_infos[agent] = infos_tz[agent_thermal_zone]
                
            # Transform the observation in a numpy array to meet the condition expected in a RLlib Environment
            agents_obs[agent] = np.array(list(obs_tz[agent_thermal_zone].values()), dtype='float32')
            # if apply, add the actuator state.
            if self.env_config['use_actuator_state']:
                agents_obs[agent] = np.concatenate(
                    (
                        agents_obs[agent],
                        [self.agent_actions[agent]],
                    ),
                    dtype='float32'
                )
            # if apply, add the agent indicator.
            if self.env_config['use_agent_indicator']:
                agents_obs[agent] = np.concatenate(
                    (
                        agents_obs[agent],
                        [agent_indicator],
                    ),
                    dtype='float32'
                )
            # if apply, add the agent type.
            if self.env_config['use_agent_type']:
                agents_obs[agent] = np.concatenate(
                    (
                        agents_obs[agent],
                        [agent_type],
                    ),
                    dtype='float32'
                )
        
        # Set the agents observation and infos to communicate with the EPEnv.
        self.infos_queue.put(agents_infos)
        self.infos_event.set()
        self.obs_queue.put(agents_obs)
        self.obs_event.set()

    def _collect_first_obs(self, state_argument):
        """
        This method is used to collect only the first observation of the environment when the episode beggins.

        Args:
            state_argument (c_void_p): EnergyPlus state pointer. This is created with `api.state_manager.new_state()`.
        """
        if self.first_observation:
            self._collect_obs(state_argument)
            self.first_observation = False
        else:
            return

    def _send_actions(self, state_argument):
        """EnergyPlus callback that sets actuator value from last decided action
        """
          
        # To not perform actions if the callbacks and the warming period are not complete.
        if not self._init_callback(state_argument):
            return
        # To not perform actions when the episode is ended or is the first timestep
        # and there are not observations.
        if self.simulation_complete:
            return
        
        # If is the first timestep, obtain the first observation before to consult for an action
        if self.first_observation:
            self.env_config['num_time_steps_in_hour'] = api.exchange.num_time_steps_in_hour(state_argument)
            self._collect_first_obs(state_argument)
            
        # Wait for an action.
        event_flag = self.act_event.wait(120)
        if not event_flag:
            print('The time waiting an action was over.')
            return
        
        # Get the central action from the EnergyPlus Environment `step` method.
        # In the case of simple agent a int value and for multiagents a dictionary.
        dict_action = self.action_fn.transform_action(self.act_queue.get())
            
        # Perform the actions in EnergyPlus simulation.       
        for agent in self._agent_ids:
            api.exchange.set_actuator_value(
                state=state_argument,
                actuator_handle=self.actuator_handles[agent],
                actuator_value=dict_action[agent]
            )
       
    def _init_callback(self, state_argument) -> bool:
        """
        Initialize EnergyPlus handles and checks if simulation runtime is ready.
        
        Args:
            state_argument (c_void_p): EnergyPlus state pointer. This is created with `api.state_manager.new_state()`.

        Returns:
            bool: True if the simulation is ready to perform actions.
        """
        self.init_handles = self._init_handles(state_argument)
        self.initialized = self.init_handles \
            and not api.exchange.warmup_flag(state_argument)
        return self.initialized

    def _init_handles(self, state_argument):
        """
        Initialize sensors/actuators handles to interact with during simulation.
        
        Args:
            state_argument (c_void_p): EnergyPlus state pointer. This is created with `api.state_manager.new_state()`.

        Returns:
            bool: True if all handles were initialized successfully.
        """
        if not self.init_handles:
            if not api.exchange.api_data_fully_ready(state_argument):
                return False
            
            if self.env_config['ep_environment_variables']:
                self.var_handles = {
                    key: api.exchange.get_variable_handle(state_argument, *var)
                    for key, var in self.variables.items()
                }
            if self.env_config['ep_thermal_zones_variables']:
                for thermal_zone in self._thermal_zone_ids:
                    self.thermal_zone_var_handles[thermal_zone].update({
                        key: api.exchange.get_variable_handle(state_argument, *var)
                        for key, var in self.thermal_zone_variables[thermal_zone].items()
                    }),
            if self.env_config['ep_object_variables']:
                for thermal_zone in self._thermal_zone_ids:
                    self.object_var_handles[thermal_zone].update({
                        key: api.exchange.get_variable_handle(state_argument, *var)
                        for key, var in self.object_variables[thermal_zone].items()
                    }),
            if self.env_config['ep_meters']:
                self.meter_handles = {
                    key: api.exchange.get_meter_handle(state_argument, meter)
                    for key, meter in self.meters.items()
                }
            self.actuator_handles = {
                key: api.exchange.get_actuator_handle(state_argument, *actuator)
                for key, actuator in self.actuators.items()
            }
            for thermal_zone in self._thermal_zone_ids:
                for handles in [
                    self.var_handles,
                    self.thermal_zone_var_handles[thermal_zone],
                    self.object_var_handles[thermal_zone],
                    self.meter_handles,
                    self.actuator_handles
                ]:
                    if any([v == -1 for v in handles.values()]):
                        available_data = api.exchange.list_available_api_data_csv(state_argument).decode('utf-8')
                        ValueError(
                            f"got -1 handle, check your var/meter/actuator names:\n"
                            f"> variables: {self.var_handles}\n"
                            f"> thermal zone variables: {self.thermal_zone_var_handles}\n"
                            f"> object variables: {self.object_var_handles}\n"
                            f"> meters: {self.meter_handles}\n"
                            f"> actuators: {self.actuator_handles}\n"
                            f"> available EnergyPlus API data: {available_data}"
                        )
                        return False
            
            self.init_handles = True
        return True

    def stop(self) -> None:
        """
        Method to stop EnergyPlus simulation and joint the threads.
        """
        if not self.simulation_complete:
            self.simulation_complete = True
        sleep(130)
        self._flush_queues()
        self.energyplus_exec_thread.join()
        self.energyplus_exec_thread = None
        self.first_observation = True
        api.runtime.clear_callbacks()
        api.state_manager.delete_state(self.energyplus_state)

    def failed(self) -> bool:
        """
        This method tells if a EnergyPlus simulations was finished successfully or not.

        Returns:
            bool: Boolean value of the success of the simulation.
        """
        return self.sim_results != 0

    def make_eplus_args(self) -> List[str]:
        """
        Make command line arguments to pass to EnergyPlus.
        
        Return:
            List[str]: List of arguments to pass to EnergyPlusEnv.
        """
        eplus_args = ["-r"] if self.env_config.get("csv", False) else []
        eplus_args += [
            "-w",
            self.env_config["epw_path"],
            "-d",
            f"{self.env_config['output_path']}/episode-{self.episode:08}",
            self.env_config["epjson_path"]
        ]
        return eplus_args
    
    def _flush_queues(self):
        """
        Method to liberate the space in the different queue objects.
        """
        for q in [self.obs_queue, self.act_queue, self.infos_queue]:
            while not q.empty():
                q.get()



"""
Action Function Base Class
==========================

This module contain the base class to create action transformer functions.
Action transformer functions are used to transform the actions of the agents.
They are used in the EnvConfig class to create the environment.
"""
from typing import Dict, Set, Any

class ActionFunction:
    """
    Base class to create action transformer functions.
    """
    def __init__(
        self,
        action_fn_config: Dict[str,Any] = {}
    ):
        """
        This class is used to transform the actions of the agents before applying
        them in the environment.

        Args:
            agents_config (Dict): This tale the configuration of the agents in the EnvConfig class.
            _agent_ids (Set): Agent ids in the environment.
        """
        self.action_fn_config = action_fn_config
    
    def transform_action(self, action:Dict[str,float]) -> Dict[str, Any]:
        """
        This method is used to transform the actions of the agents before applying.

        Args:
            action (Dict[str,float]): Action provided by the policy.

        Returns:
            Dict[str, Any]: A dict of transformed action for each agent in the environment.
        """
        return action



"""
Environment Utilities
=====================

This module contain the methods used in the environment process.
"""

from typing import Tuple, Dict, Any, List, Set, Optional
from gymnasium.spaces import Box, Discrete
import numpy as np

def EP_API_add_path(version:Optional[str]="23-2-0", path:Optional[str]=None):
    """
    This method add the EnergyPlus Python API to the system path. This allow to use the 
    EnergyPlus program in Python. The minimal version of EnergyPlus is 9.3.0 and the default
    version (and the stable one) for eprllib is 23-2-0.

    Args:
        version (str, optional): Numeric version of EnergyPlus. Defaults to "23-2-0".
        path (Optional[str], optional): Complete path to the EnergyPlus installation directory 
        if this is different that the default installation. Defaults to None.
    """
    import sys
    
    if path is not None:
        sys.path.insert(0, f"{path}")
    else:  
        os_platform = sys.platform
        if os_platform == "linux":    
            sys.path.insert(0, f"/usr/local/EnergyPlus-{version}")
        else:
            sys.path.insert(0, f"C:/EnergyPlusV{version}")
    
    return print(f"EnergyPlus API path added: {sys.path[0]}")
    
def actuators_to_agents(agent_config:Dict[str, List]) -> Tuple[List,List,Dict,Dict]:
    """
    Take the ep_actuator dict and transform it to the agent, thermal zone, and actuator type dict.

    Args:
        agent_config (Dict): ep_actuator dict in the env_config.

    Returns:
        Tuple[Dict,Dict,Dict,Dict]: agent, thermal zone, and actuator type.
    """
    # get the keys and values from the ep_actuator definitions
    keys = agent_ids = list(agent_config.keys())
    values = list(agent_config.values())
    
    # Verify that all the agents have the same quantity of attributes
    for i in range(len(values)):
        assert len(values[0])==len(values[i])
    
    # compose the actuators for EnergyPlus handle
    actuator = []
    for i in range(len(keys)):
        actuator_tuple = (values[i][0], values[i][1], values[i][2])
        actuator.append(actuator_tuple)
    # asign the actuator handle config to the agent
    agents_actuators = {}
    for i in range(len(keys)):
        agents_actuators[keys[i]] = actuator[i]

    # identify the thermal zones where there are agents
    agent_thermal_zone_names = []
    for i in range(len(keys)):
        agent_thermal_zone_names.append(values[i][3])
    # asign the thermal zone to each agent
    agents_thermal_zones = {}
    for i in range(len(keys)):
        agents_thermal_zones[keys[i]] = agent_thermal_zone_names[i]
    # a list with the thermal zones with agents (without repeat)
    thermal_zone_ids = []
    for zone in agent_thermal_zone_names:
        if not zone in  thermal_zone_ids:
            thermal_zone_ids.append(zone)
    
    # agent type
    typ = []
    for i in range(len(keys)):
        typ.append(values[i][4])
    agents_types = {}
    for i in range(len(keys)):
        agents_types[keys[i]] = typ[i]
    
    agents_str = ", ".join(agent_ids)
    print(f"The environment is defined with {len(agent_ids)} agents: {agents_str}")
    
    return agent_ids, thermal_zone_ids, agents_actuators, agents_thermal_zones, agents_types

def obs_space(env_config:Dict, _thermal_none_ids:Set):
    """
    This method construct the observation space of the environment.

    Args:
        env_config (Dict): The environment configuration dictionary.

    Returns:
        space.Box: The observation space of the environment.
    """
    
    obs_space_len = 0
    
    # actuator state.
    if env_config['use_actuator_state']:
        obs_space_len += 1
        
    # agent_indicator.
    if env_config['use_agent_indicator']:
        obs_space_len += 1
        
    # agent type.
    if env_config['use_agent_type']:
        obs_space_len += 1
        
    # building properties.
    if env_config['use_building_properties']:
        for thermal_zone in _thermal_none_ids:
            thermal_zone_name = thermal_zone
            break
        obs_space_len += len([key for key in env_config['building_properties'][thermal_zone_name].keys()])
        
    # weather prediction.
    if env_config['use_one_day_weather_prediction']:
        obs_space_len += 24*6
        
    # variables and meters.
    if env_config['ep_environment_variables']:
        obs_space_len += len(env_config['ep_environment_variables'])
    
    if env_config['ep_thermal_zones_variables']:
        obs_space_len += len(env_config['ep_thermal_zones_variables'])
    
    if env_config['ep_object_variables']:
        for thermal_zone in _thermal_none_ids:
            thermal_zone_name = thermal_zone
            break
        obs_space_len += len([key for key in env_config['ep_object_variables'][thermal_zone_name].keys()])
        
    if env_config['ep_meters']:
        obs_space_len += len(env_config['ep_meters'])
        
    if env_config['time_variables']:
        obs_space_len += len(env_config['time_variables'])
        
    if env_config['weather_variables']:
        obs_space_len += len(env_config['weather_variables'])
        
    # discount the not observable variables.
    if env_config['no_observable_variables']:
        for thermal_zone in _thermal_none_ids:
            thermal_zone_name = thermal_zone
            break
        obs_space_len -= len(env_config['no_observable_variables'][thermal_zone_name])
        
    # construct the observation space.
    return Box(float("-inf"), float("inf"), (obs_space_len,))

def continuous_action_space():
    """
    This method construct the action space of the environment.
    
    Returns:
        gym.Box: Continuous action space with limits between [0,1].
    """
    return Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)

def discrete_action_space():
    """
    This method construct the action space of the environment.
    
    Returns:
        gym.Discrete: Discrete action space with limits between [0,10] and a step of 1.
    """
    return Discrete(11)

def environment_variables(env_config: Dict[str, Any]) -> Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]:
    """
    The EnergyPlus outdoor environment variables are defined in the environment configuration.

    Args:
        env_config (Dict[str, Any]): The EnvConfig dictionary.

    Returns:
        Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]: The environment variables and their handles.
    """
    variables:Dict[str, Tuple [str, str]] = {}
    if env_config['ep_environment_variables']:
        variables = {variable: (variable, 'Environment') for variable in env_config['ep_environment_variables']}
    
    var_handles: Dict[str, int] = {}
    
    return variables, var_handles

def thermal_zone_variables(env_config: Dict[str, Any], _thermal_zone_ids:Set) -> Tuple[Dict[str, Dict[str, Tuple[str,str]]],Dict[str,int]]:
    """
    The EnergyPlus thermal zone variables are defined in the environment configuration.

    Args:
        env_config (Dict[str, Any]): The EnvConfig dictionary.

    Returns:
        Tuple[Dict[str, Dict[str, Tuple [str, str]]],Dict[str,int]]: The thermal zone variables and their handles.
    """
    thermal_zone_variables: Dict[str, Tuple [str, str]] = {thermal_zone: {} for thermal_zone in _thermal_zone_ids}
    if env_config['ep_thermal_zones_variables']:
        for thermal_zone in _thermal_zone_ids:
            thermal_zone_variables[thermal_zone].update({variable: (variable, thermal_zone) for variable in env_config['ep_thermal_zones_variables']})
    thermal_zone_var_handles: Dict[str, int] = {thermal_zone: {} for thermal_zone in _thermal_zone_ids}
    
    return thermal_zone_variables, thermal_zone_var_handles

def object_variables(env_config: Dict[str, Any], _thermal_zone_ids: Set) -> Tuple[Dict[str, Dict[str, Tuple[str,str]]],Dict[str,int]]:
    """
    The EnergyPlus object variables are defined in the environment configuration.

    Args:
        env_config (Dict[str, Any]): The EnvConfig dictionary.

    Returns:
        Tuple[Dict[str, Dict[str, Tuple [str, str]]],Dict[str,int]]: The object variables and their handles.
    """
    object_variables: Dict[str, Dict[str, Tuple [str, str]]] = {}
    if env_config['ep_object_variables']:
        object_variables = env_config['ep_object_variables']
    object_var_handles = {thermal_zone: {} for thermal_zone in _thermal_zone_ids}
    
    return object_variables, object_var_handles

def meters(env_config: Dict[str, Any]) -> Tuple[Dict[str, str], Dict[str,int]]:
    """The EnergyPlus meters are defined in the environment configuration.
    
    Args:
        env_config (Dict[str, Any]): The EnvConfig dictionary.

    Returns:
        Tuple[Dict[str, str], Dict[str,int]]: The meters and their handles.
    """
    meters: Dict[str,str] = {}
    if env_config['ep_meters']:
        meters = {key: key for key in env_config['ep_meters']}
    meter_handles: Dict[str, int] = {}
    
    return meters, meter_handles

def actuators(env_config: Dict[str, Any], _agent_ids:Set) -> Tuple[Dict[str,Tuple[str,str,str]], Dict[str,int]]:
    """The EnergyPlus actuators are defined in the environment configuration.

    Args:
        env_config (Dict[str, Any]): The EnvConfig dictionary.

    Returns:
        Tuple[Dict[str,Tuple[str,str,str]], Dict[str,int]]: The actuators and their handles.
    """
    actuators: Dict[str,Tuple[str,str,str]] = {agent: env_config['agents_config'][agent]['ep_actuator_config'] for agent in _agent_ids}
    actuator_handles: Dict[str, int] = {}
    
    return actuators, actuator_handles


"""
Henderson et al. (2024)
=======================

This module contains the reward function class for the EnergyPlusEnv_v0 environment.
The reward function is based on the Henderson et al. (2024) paper, which proposes a reward function
for optimizing the energy consumption and comfort of a building. The reward function is divided into
two components: energy demand penalty and comfort penalty. The energy demand penalty is based on the
maximum energy demand for the entire episode, and the comfort penalty is based on the average PPD comfort
metric for the entire episode. The reward function is normalized by dividing each term by the maximum
value for the entire episode, and multiplying by a ponderation factor for the energy and (1-beta) for
the comfort. Both terms are negatives, representing a penalti for demand energy and for generate discomfort.

The reward function is designed to be used in conjunction with the EnergyPlusEnv_v0 environment, which
provides the necessary information for calculating the reward.
"""

import numpy as np
from typing import Dict, Any
from eprllib.RewardFunctions.RewardFunctions import RewardFunction

class henderson_2024(RewardFunction):
    def __init__(
        self,
        reward_fn_config: Dict[str,Any]
        ) -> Dict[str,float]:
        """
        This class implements the Henderson et al. (2024) reward function for the EnergyPlusEnv_v0 environment.
        The reward function is divided into two components: energy demand penalty and comfort penalty. The energy 
        demand penalty is based on the maximum energy demand for the entire episode, and the comfort penalty is 
        based on the average PPD comfort metric for the entire episode. The reward function is normalized by 
        dividing each term by the maximum value for the entire episode, and multiplying by a ponderation factor 
        for the energy and (1-beta) for the comfort. Both terms are negatives, representing a penalti for demand 
        energy and for generate discomfort.

        Args:
            reward_fn_config (Dict[str,Any]): The dictionary is to configurate the variables that use each agent
            to calculate the reward. The dictionary must to have the following keys: agent_thermal_zone_ids, comfort_reward,
            energy_reward, cut_reward_len_timesteps, beta, ppd_name, T_interior_name, occupancy_name, cooling_energy_ref,
            heating_energy_ref, cooling_name, heating_name. All this variables start with the name of the agent and then
            the valuo of the reference name.

        Returns:
            Dict[str,float]: The reward value for each agent in the timestep.
        """
        super().__init__(reward_fn_config)
        
        self.thermal_zone: Dict[str,str] = reward_fn_config["thermal_zone"]
        self.comfort_reward: Dict[str,bool] = reward_fn_config['comfort_reward']
        self.energy_reward: Dict[str,bool] = reward_fn_config['energy_reward']
        self.cut_reward_len_timesteps: Dict[str,float] = reward_fn_config['cut_reward_len_timesteps']
        self.beta: Dict[str,float] = reward_fn_config['beta']
        self.ppd_name: Dict[str,str] = reward_fn_config['ppd_name']
        self.T_interior_name: Dict[str,str] = reward_fn_config['T_interior_name']
        self.occupancy_name: Dict[str,str] = reward_fn_config['occupancy_name']
        self.cooling_energy_ref: Dict[str,str] = reward_fn_config['cooling_energy_ref']
        self.heating_energy_ref: Dict[str,str] = reward_fn_config['heating_energy_ref']
        self.cooling_name: Dict[str,str] = reward_fn_config['cooling_name']
        self.heating_name: Dict[str,str] = reward_fn_config['heating_name']
        
        # dictionary to save the values of each agent when the reward is dalyed.
        self.ppd_dict = {key: [] for key in self.thermal_zone.keys()}
        self.energy_dict = {key: [] for key in self.thermal_zone.keys()}
        
        self.timestep = 0
        
    def calculate_reward(
    self,
    infos: Dict[str,Dict[str,Any]] = None
    ) -> Dict[str,float]:
        """This function returns the normalize reward calcualted as the sum of the penalty of the energy 
        amount of one week divide per the maximun reference energy demand and the average PPD comfort metric
        divide per the maximal PPF value that can be take (100). Also, each term is divide per the longitude
        of the episode and multiply for a ponderation factor of beta for the energy and (1-beta) for the comfort.
        Both terms are negatives, representing a penalti for demand energy and for generate discomfort.

        Args:
            self (Environment): RLlib environment.
            obs (dict): Zone Mean Air Temperature for the Thermal Zone in °C.
            infos (dict): infos dict must to provide the occupancy level and the Zone Mean Temperature.

        Returns:
            float: reward normalize value
        """
        self.timestep += 1
        # dictionary to save the reward values of each agent.
        reward_dict = {key: 0. for key in self.thermal_zone.keys()}
        
        # asign the ThermalZone values above defined to each agent
        for agent in self.thermal_zone.keys():
            if self.comfort_reward[agent]:
                ppd = infos[agent][self.ppd_name[agent]]
                occupancy = infos[agent][self.occupancy_name[agent]]
                T_interior = infos[agent][self.T_interior_name[agent]]
                if occupancy > 0:
                    if infos[agent][self.ppd_name[agent]] < 5:
                        ppd = infos[agent][self.ppd_name[agent]] = 5
                    else:
                        ppd = infos[agent][self.ppd_name[agent]]
                else:
                    ppd = infos[agent][self.ppd_name[agent]] = 0
                # If there are not people, only the reward is calculated when the environment is far away
                # from the comfort temperature ranges. This limits are recommended in EnergyPlus documentation:
                # InputOutput Reference p.522
                if T_interior > 29.4:
                    ppd = infos[agent][self.ppd_name[agent]] = 100
                elif T_interior < 16.7:
                    ppd = infos[agent][self.ppd_name[agent]] = 100
                    
                self.ppd_dict[agent].append(ppd)
            
            if self.energy_reward[agent]:
                cooling_meter = infos[agent][self.cooling_name[agent]]
                heating_meter = infos[agent][self.heating_name[agent]]
                
                self.energy_dict[agent].append(cooling_meter/self.cooling_energy_ref[agent]+heating_meter/self.heating_energy_ref[agent])
        
        # calculate the reward if the timestep is divisible by the cut_reward_len_timesteps.
        # if don't return 0.
        for agent in self.thermal_zone.keys():
            if self.timestep % self.cut_reward_len_timesteps[agent] == 0:
                if self.comfort_reward[agent]:
                    ppd_avg = sum(self.ppd_dict[agent])/len(self.ppd_dict[agent])
                    rew1 = -(1-self.beta[agent])*(1/(1+np.exp(-0.1*(ppd_avg-45))))
                else:
                    rew1 = 0
                
                if self.energy_reward[agent]:
                    rew2 = -self.beta[agent]*(sum(self.energy_dict[agent])/len(self.energy_dict[agent]))
                else:
                    rew2 = 0
                
                reward_dict[agent] = rew1 + rew2
                
        # emptly the lists
        for agent in self.thermal_zone.keys():
            if self.timestep % self.cut_reward_len_timesteps[agent] == 0:
                self.ppd_dict = {key: [] for key in self.thermal_zone.keys()}
                self.energy_dict = {key: [] for key in self.thermal_zone.keys()}
                
        return reward_dict



"""
eprllib
=======

The package `eprllib` provides a set of tools for the development of energy performance
research in the context of the Python programming language. The package is structured
as follows:

.. toctree::
   :maxdepth: 1
   :caption: Contents:
   
   ActionFunctions
   Agents
   Env
   EpisodeFunctions
   ExampleFiles
   PostProcess
   PreProcess
   RewardFunctions
   Tools

"""


"""
Reward Funtion
==============

This module contains the base reward function.

It has been preferred to use the `infos` dictionary and not the observation, since the latter is 
a numpy array and cannot be called by key values, which is prone to errors when developing the program 
and indexing a arrangement may change.
"""

from typing import Dict, Any

class RewardFunction:
    """
    This class is the base class for the reward function.
    """
    def __init__(
        self,
        reward_fn_config: Dict[str,Any]
    ):
        self.reward_config = reward_fn_config
    
    def calculate_reward(
        self,
        infos: Dict[str,Dict[str,Any]]
        ) -> Dict[str,float]:
        """
        This method must be implemented in the subclass.

        Args:
            infos (Dict[str,Dict[str,Any]]): The infos dictionary containing the necessary information for calculating the reward.

        Returns:
            Dict[str,float]: The calculated reward as a dictionary with the keys 'agent'.
        """
        return NotImplementedError("This method must be implemented in the subclass.")



"""
Conventional Agent
==================

This module contains the base class of the conventional agents.
"""
from typing import Dict, Any

class ConventionalAgent:
    
    def __init__(
        self,
        config: Dict[str,Any],
        **kargs,
    ):
        """
        This agent perform conventional actions in an EnergyPlus model based on fixed rules
        that take into account the basics variables as temperature, radiation, humidity and others.
        
        Args:
            config (Dict[str,Any]): Here you can set the configuration of the agent.
        """
        self.config = config
    
    def compute_single_action(self, infos:Dict[str, int|float], **kargs) -> int|float:
        """Implement here your own function."""
        raise NotImplementedError
